// ****************************************************************************
// BarTwister effect (c) Jarmo
// This code is public domain. Use at your own risk.
// http://z0b.kapsi.fi
// ****************************************************************************

#include <SDL/SDL.h>
#include <algorithm>
#include <cmath>

// animation speed, change if you want
static const float fps = 60.0f, speed = 4.0f;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// EFFECT CODE

// effect and texture sizes, in pixels (also the window size)
static const uint32_t effectW = 384, effectH = 192,
                      textureW = 384, textureH = 64;

typedef float (*equationFunction)(const float angle, const uint32_t x);

static inline int ftoi(const float f)
{
    return f >= 0.0f ? (int)(f + 0.5f) : (int)(f - 0.5f);
}

// draw one shaded slice of texture
static void drawSideTex(uint8_t *dest,
                        const uint8_t *texture,
                        const int x,
                        const uint32_t y1, const uint32_t y2,
                        const float z1, const float z2)
{
    if (!dest || !texture)
        return;

    const uint32_t h = y2 - y1,
                   tw = textureW * 3U;

    if (h < 1U || h > effectH - 1U)
        return;

    // During profiling, I noticed that floating-point texture interpolation with
    // fixed-point shading interpolation is the fastest combination. I don't know why.

    // texture interpolation (+ 2U is a hack to prevent small seams)
    const float dT = ((float)textureH / (h + 2U));
    float tex = 0.0f;

    // shading interpolation (fixed-point)
    // (there are many ways to calculate s1 and s2, this one looks good)
    const int s1 = ftoi(255.0f * std::min(z1, z2)),
              s2 = ftoi(255.0f * std::max(z1, z2));

    const int dS = ((s2 - s1) << 16) / (signed)h;
    int shade = s1 << 16;

    // setup pointers
    const uint8_t *column = texture + (x * 3);
    uint8_t *destination = dest + ((y1 * effectW + x) << 2);
    const uint32_t stride = (effectW << 2) - 3;
    uint32_t y = y1;

    // draw
    while (y <= y2) {
        const uint8_t *source = column + ((uint32_t)(tex + 0.5f) * tw);

        *destination++ = (source[2] * (shade >> 16)) >> 8;
        *destination++ = (source[1] * (shade >> 16)) >> 8;
        *destination++ = (source[0] * (shade >> 16)) >> 8;

        destination += stride;
        tex += dT;
        shade += dS;
        ++y;
    }
}

// renders the effect, does not validate texture size or anything
static void barTwister(uint8_t *output, const uint8_t *texture, const float angle,
                       const equationFunction equation)
{
    // (Y, Z) source coordinates for X-aligned quads (spi4=sin(PI/4); when
    // rotated, it gives values between -1 and 1). Scale must be below 1.0.
    static const float spi4 = 0.7071067811865f,
                       scale = 0.95f;

    static const float coords[4][2] =
    {
        { -spi4 * scale,  spi4 * scale },
        { -spi4 * scale, -spi4 * scale },
        {  spi4 * scale, -spi4 * scale },
        {  spi4 * scale,  spi4 * scale }
    };

    // Y and Z for each vertex after rotation and scaling
    uint32_t y[4];
    float z[4];

    if (!output)
        return;

    memset(output, 0, effectW * effectH * 4 * sizeof(uint8_t));

    if (!texture || !equation)
        return;

    const float height = effectH - 1;

    // draw each slice
    for (uint32_t x = 0U; x < effectW; x++) {
        // get rotation angle
        const float r = equation(angle, x),
                    c = cos(r), s = sin(r);

        // rotate vertices
        for (uint32_t i = 0U; i < 4U; i++) {
            z[i] =            ((c * coords[i][1] - s * coords[i][0]) + 1.0f) / 2.0f;
            y[i] = (uint32_t)(((s * coords[i][1] + c * coords[i][0]) / 2.0f + 0.5f) * height + 0.5f);
            y[i] = std::min(effectH - 1, y[i]);
        }

        // Draw sides, sort by Y to get proper ordering. We always draw from top
        // to bottom (Y1 > Y2); if (Y1 < Y2), then the side is back-facing.
        for (uint32_t now = 0U; now < 4U; now++) {
            const uint32_t next = (now + 1U) & 3U;

            if (y[now] > y[next])
                drawSideTex(output, texture, x, y[next], y[now], z[next], z[now]);
        }
    }
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// EQUATIONS

static float rotate(const float angle, const uint32_t x)
{
    return angle / 128.0f;
}

static float sine2(const float angle, const uint32_t x)
{
    return x / (float)textureW * 2.0f + angle / 128.0f;
}

static float sine3(const float angle, const uint32_t x)
{
    return x / (float)textureW * 3.0f + angle / 128.0f;
}

static float simple(const float angle, const uint32_t x)
{
    return sin((float)(angle + x) / 64.0f);
}

static float twisty(const float angle, const uint32_t x)
{
    return ((x + 64) * sin(((angle + x) / 5.0f) * 0.017453292f) / 1.5f) * 0.017453292f;
}

#define NUM_EQUATIONS   5

// all equations
static const equationFunction equations[NUM_EQUATIONS] =
{
    rotate,
    sine2,
    sine3,
    simple,
    twisty,
};

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// MAIN

int main(int argc, char *argv[])
{
    // init SDL
    if (SDL_Init(SDL_INIT_VIDEO)) {
        fprintf(stderr, "SDL init failed: %s\n", SDL_GetError());
        return 1;
    }

    SDL_Surface *screen = SDL_SetVideoMode(effectW, effectH, 32, SDL_SWSURFACE);

    if (!screen) {
        fprintf(stderr, "Unable to set 640x480 video: %s\n", SDL_GetError());
        SDL_Quit();
        return 1;
    }

    SDL_WM_SetCaption("BarTwister", NULL);

    // load the texture
    uint8_t *texture = new(std::nothrow) uint8_t[textureW * textureH * 3];

    if (!texture) {
        fprintf(stderr, "No memory for the texture\n");
        SDL_Quit();
        return 1;
    }

    FILE *f = fopen("texture.raw", "rb");

    if (!f) {
        fprintf(stderr, "Can't open the texture file\n");
        delete [] texture;
        SDL_Quit();
        return 1;
    }

    fread(texture, 3, textureW * textureH, f);
    fclose(f);

    // main loop
    uint32_t prev = SDL_GetTicks();
    bool done = false;
    float angle = 0.0f;
    int equation = 2;

    while (!done) {
        if (SDL_MUSTLOCK(screen))
            if (!SDL_LockSurface(screen)) {
                fprintf(stderr, "SDL_LockSurface() failed: %s\n", SDL_GetError());
                break;
            }

        // draw
        barTwister((uint8_t *)screen->pixels, texture, angle, equations[equation]);

        // spin
        const uint32_t now = SDL_GetTicks();

        angle += speed * ((float)(now - prev) * fps / 1000.0f);
        prev = now;

        if (SDL_MUSTLOCK(screen))
            SDL_UnlockSurface(screen);

        SDL_UpdateRect(screen, 0, 0, effectW, effectH);

        // handle events, if any
        SDL_Event event;

        while (SDL_PollEvent(&event)) {
            switch (event.type) {
                case SDL_KEYDOWN:
                {
                    const int sym = event.key.keysym.sym;

                    if (sym == SDLK_ESCAPE)
                        done = true;
                    else if (sym >= SDLK_1 && sym < (SDLK_1 + NUM_EQUATIONS))
                        equation = sym - SDLK_1;

                    break;
                }

                case SDL_QUIT:
                    done = true;
                    break;
            }
        }
    }

    delete [] texture;
    SDL_Quit();
    return 0;
}
